from fastapi import APIRouter, HTTPException
from fastapi.responses import FileResponse
from ManagerAPI.api.managers.job_manager import JobManager
from ManagerAPI.api.models.managedJob import ManagedJobModel
from ManagerAPI.api.managers.state_manager import StateManager
from ManagerAPI.api.managers.connection_manager import RedisConnectionConfig
import logging

log = logging.getLogger('uvicorn')

jobRouter = APIRouter()
jobManager = JobManager(RedisConnectionConfig(), StateManager())


@jobRouter.get('/jobs', tags=['jobs'])
async def get_jobs():
    queued_jobs = {}

    for queue in jobManager.stateManager.states():
        log.debug(f'Getting jobs from queue {queue} - {jobManager.stateManager[queue]}')
        queued_jobs[jobManager.stateManager[queue]] = jobManager.get_jobs_from_state(queue=queue)

    return queued_jobs


@jobRouter.get('/job/id/{job_id}', tags=['jobs'])
async def get_job(job_id):

    job = jobManager.get_job(job_id=job_id)
    if job is None:
        log.error(f'Job {job_id} not found')
        raise HTTPException(status_code=404, detail='Job not found')

    return job.model_dump()


@jobRouter.post('/job', tags=['jobs'])
async def update_job(job: ManagedJobModel):

    log.info(f'Updating job: {job.model_dump()}')

    if not jobManager.update_job(job=job):
        log.error(f'Failed to update job {job.job_id}')
        raise HTTPException(status_code=500, detail='Failed to update job')

    return job.model_dump()


@jobRouter.delete('/job/id/{job_id}', tags=['jobs'])
async def delete_job(job_id, force: bool = False):

    if not jobManager.get_job(job_id=job_id):
        log.error(f'Job {job_id} not found')
        raise HTTPException(status_code=404, detail='Job not found')

    if not jobManager.delete_job(job_id=job_id, force=force):
        log.error(f'Failed to delete job {job_id}')
        raise HTTPException(status_code=500, detail='Failed to delete job from Redis')

    return {'Deleted': job_id}


@jobRouter.get('/job/queue/{queue}', tags=['jobs'])
async def get_jobs_from_queue(queue):

    job_id = jobManager.get_job_from_queue(queue=queue)

    if job_id is None:
        log.error(f'No jobs in queue {queue}')
        raise HTTPException(status_code=404, detail='No jobs in queue')

    return {queue: job_id}


@jobRouter.get('/job/id/{job_id}/status', tags=['jobs'])
async def get_job_status(job_id):

    job = jobManager.get_job(job_id=job_id)

    if job is None:
        log.error(f'Job {job_id} not found')
        raise HTTPException(status_code=404, detail='Job not found')

    return {'status': job.job_status}


@jobRouter.get('/job/id/{job_id}/video', tags=['jobs'])
async def get_job_video(job_id):

    job = jobManager.get_job(job_id=job_id)

    if job is None:
        log.error(f'Job {job_id} not found')
        raise HTTPException(status_code=404, detail='Job not found')

    if job.job_status == "error":
        log.error(f'Job {job_id} errored before video creation')
        raise HTTPException(status_code=404, detail='Job errored before video creation')

    if job.job_status not in ["packed", "uploaded", "completed"]:
        log.error(f'Job {job_id} video not packed')
        raise HTTPException(status_code=404, detail='Video not created yet')

    if not job.video_urn:
        log.error(f'Job {job_id} video not found')
        raise HTTPException(status_code=404, detail='Video is expected, but not found')

    return FileResponse(job.video_urn)



from redis import StrictRedis, exceptions
from ManagerAPI.api.models.managedJob import ManagedJobModel
import logging
import shutil
import os

log = logging.getLogger('uvicorn')


class JobManager:

    def __init__(self, redisConnectionManagerConfig, stateManager):
        self.redis = StrictRedis(host=redisConnectionManagerConfig.host, port=redisConnectionManagerConfig.port, db=redisConnectionManagerConfig.db,
                                 decode_responses=redisConnectionManagerConfig.decode_responses)
        self.stateManager = stateManager

    def update_job(self, job: ManagedJobModel):

        try:
            self.redis.hmset(job.job_id, job.model_dump())
        except exceptions.ConnectionError as e:
            log.error(f'Redis connection issue while updating {job.job_id} : {e}')
            return False

        log.info(f'Adding job {job.job_id} to Redis')

        return self.post_job_to_queue(queue=job.job_status, job_id=job.job_id)

    def get_job(self, job_id: str):

        try:
            job = self.redis.hgetall(job_id)
        except exceptions.ConnectionError as e:
            log.error(f'Redis connection issue while fetching {job_id} : {e}')
            return None

        if not job:
            log.error(f'Job {job_id} not found')
            return None

        log.debug(f'Fetched Job {job_id} : {job}')

        return ManagedJobModel(**job)

    '''
      Gets a list of job ids from a given queue.
    '''
    def get_jobs_from_state(self, queue: str):
        jobs = []

        '''
        Check if queue is valid
        '''
        if not self.stateManager.valid_state(queue):
            return None

        try:
            jobs = self.redis.lrange(self.stateManager[queue], 0, -1)
        except exceptions.ConnectionError as e:
            log.error(f'Redis connection issue while fetching {queue} : {e}')

        return jobs

    def post_job_to_queue(self, queue: str, job_id: str):

        '''
        Check if queue is valid
        '''
        if not self.stateManager.valid_state(queue):
            log.error(f'Invalid queue {queue}')
            return False

        '''
        Check if job is already in queue
        '''
        if job_id in self.get_jobs_from_state(queue=queue):
            log.info(f'Job {job_id} already in queue {queue}')
            return True

        try:
            self.redis.rpush(self.stateManager[queue], job_id)
        except exceptions.ConnectionError:
            return False

        log.info(f'Added job {job_id} to queue {self.stateManager[queue]}')

        return True

    def get_job_from_queue(self, queue: str):

        if not self.stateManager.valid_state(queue):
            return None

        try:
            job_id = self.redis.lpop(self.stateManager[queue])
        except exceptions.ConnectionError:
            return None

        if not job_id:
            return None

        return self.get_job(job_id)

    def delete_job(self, job_id: str, force: bool = False):

        JOB_IMG_VID_DIR = os.getenv('IMAGE_VIDEO_DIR', '/tmp')

        file_path = f'{JOB_IMG_VID_DIR}/{job_id}'
        log.debug(f'Attempting to delete job {job_id} from {file_path}')

        if not force:
            if os.path.exists(file_path):
                shutil.rmtree(file_path)
            else:
                log.error(f'Job {job_id} data not found on disk')
                return False

        '''
        Add block to delete associated job files from disk
        '''
        try:
            self.redis.delete(job_id)
        except exceptions.ConnectionError:
            log.error(f'Failed to delete job {job_id}')
            return False

        log.info(f'Deleted job {job_id}')

        return True




from ManagerAPI.api.managers.state_manager import StateManager
from ManagerAPI.api.models.managedJob import ManagedJobModel
from time import time
import requests
import logging as log
import os

BASE_JOB_HANDLER_URL = os.getenv('MANAGERAPI_BASE_URL', 'http://localhost:8000/')


class JobHandler:

    stateManager = StateManager()
    baseConnectionURL = BASE_JOB_HANDLER_URL

    def __init__(self, job_queue: str):
        self.job_queue = job_queue

    def available_job(self) -> bool:

        try:
            jobs = requests.get(self.baseConnectionURL + "jobs").json()
        except requests.exceptions.ConnectionError:
            log.error(f'[Job Handler] Failed to fetch jobs from {self.baseConnectionURL + "jobs"} for {self.job_queue}')
            return False
        except requests.exceptions.ReadTimeout:
            log.error(f'[Job Handler] Timeout fetching jobs from {self.baseConnectionURL + "jobs"} for {self.job_queue}')
            return False

        log.debug(f'Fetched {self.stateManager[self.job_queue]} available jobs - current count {len(jobs[self.stateManager[self.job_queue]])}')

        return (jobs and len(jobs[self.stateManager[self.job_queue]]) > 0)

    def fetch_job(self):

        try:
            job_query = requests.get(self.baseConnectionURL + "job/queue/" + self.job_queue)
        except requests.exceptions.ConnectionError:
            log.error(f'[Job Handler] Failed to fetch job from {self.baseConnectionURL + "job/queue/" + self.job_queue}')
            return None
        except requests.exceptions.ReadTimeout:
            log.error(f'[Job Handler] Timeout fetching job from {self.baseConnectionURL + "job/queue/" + self.job_queue}')
            return None

        if job_query.status_code != 200:
            return None

        log.debug(f'Fetched job {job_query.json()[self.job_queue]} from {self.baseConnectionURL + "job/queue/" + self.job_queue}')

        return ManagedJobModel(**job_query.json()[self.job_queue])

    def update_job(self, job):

        log.debug(f'Attempting to update job {job.job_id} to {self.baseConnectionURL + "job"}')

        try:
            job_update = requests.post(self.baseConnectionURL + "job", json=job.model_dump())
        except requests.exceptions.ConnectionError:
            log.error(f'[Job Handler] Failed to update job {job.job_id} to {self.baseConnectionURL + "job"}')
            return False
        except requests.exceptions.ReadTimeout:
            log.error(f'[Job Handler] Timeout updating job {job.job_id} to {self.baseConnectionURL + "job"}')
            return False

        if job_update.status_code != 200:
            return False

        log.debug(f'Updated job {job.job_id} to {self.baseConnectionURL + "job"}')

        return True

    def error_job(self, job, error_message: str):

        job.job_error = error_message
        job.job_end_time = int(time())
        job.job_status = "error"

        self.update_job(job)

        log.debug(f'Error status set for job {job.job_id}')

        return True



from uuid import uuid4
from datetime import datetime
from time import time
from pydantic import BaseModel, Field, model_validator, ValidationInfo
from typing import Union, Optional


class ManagedJobModel(BaseModel):
    job_id: Optional[str] = Field(default_factory=lambda: str(uuid4()))
    job_status: Union[str, None] = "new"
    job_start_time: Optional[int] = Field(default_factory=lambda: int(time()))
    job_end_time: Union[int, None] = -1
    job_error: Union[str, None] = ''
    img_date: str
    image_links: Union[str, None] = ''
    gif_urn: Union[str, None] = ''
    video_urn: Union[str, None] = ''
    img_resolution: str
    yt_video_id: Union[str, None] = ''
    publish_public: Union[str, None] = ''
    region: str
    storm_id : Union[str,None] = ''

    class Config:
        orm_mode = True

    #Custom model validation to ensure a storm_id is set if the storm region is set
    @model_validator(mode='after')
    def check_storm_id_present_if_storm_region_set(self):
        region = self.region
        storm_id = self.storm_id

        if region == "storm" and not storm_id:
            raise ValueError("Storm Region set, but no storm id provided")
        
        return self

    def get_formatted_date(self):
        date_str = self.img_date.split("-")
        return datetime(int(date_str[0]), int(date_str[1]), int(date_str[2])).strftime("%Y%j")

    def get_title_date(self):
        date_str = self.img_date.split("-")
        return datetime(int(date_str[0]), int(date_str[1]), int(date_str[2])).strftime("%b %d %Y")


from LogHandler.logHandler import LogHandler
from JobHandler.jobHandler import JobHandler
from ManagerAPI.api.models.managedJob import ManagedJobModel
from VideoUploader.videoUploader.managers.upload_manager import YTAuthManager, YTVideoManager
from time import sleep, time
import logging as log


def get_video_type(job: ManagedJobModel):

    resolution = job.img_resolution.split("x")

    if resolution[0] == resolution[1]:
        return "shorts"

    return "normal"


def get_title(job: ManagedJobModel):
    title_base = f'{job.region.replace("_", " ").upper()} {job.get_title_date()} - 24 Hour Timelapse'
    title_base_shorts = f'{job.region.replace("_", " ").title()} {job.get_title_date()} Weather Timelapse #weathertoday'

    if get_video_type(job) == "shorts":
        return title_base_shorts

    return title_base


def publish_privacy_status(job: ManagedJobModel):
    if job.publish_public:
        return "public"

    return "private"


def get_description(job: ManagedJobModel):
    
    description = f' Timelapse of the {job.region.replace("_", " ").title()} weather for {job.get_title_date()} \n Imagery credit to NOAA / NESDIS Center for Satellite Applications and Research.'
    log.debug(f"Attached Description for {job.job_id} - Description: {description}")
    return description


def upload_video(job: ManagedJobModel):
    log.debug(f'Uploading video for job: {job.job_id}')

    ytAuthManager = YTAuthManager().bootstrap()
    ytVideoManager = YTVideoManager(title=get_title(job=job),
                                    description=get_description(job=job),
                                    video_type=get_video_type(job=job),
                                    privacy_status=publish_privacy_status(job=job),
                                    video_file=job.video_urn,
                                    auth_manager=ytAuthManager)

    video_id = ytVideoManager.upload_video(ytAuthManager.fetch_credentials())
    log.info(f'Uploaded video {video_id}')

    if not video_id:
        log.error(f'Failed to upload video for job: {job.job_id}')
        job.job_status = "error"
        job.job_error += f'Failed to upload video {job.job_id}'
        return False

    job.yt_video_id = video_id

    return video_id


def run():

    jobHandler = JobHandler("packed")

    if not jobHandler.available_job():
        log.info('No jobs available')
        return

    job = jobHandler.fetch_job()

    if not job:
        log.info('No jobs available')
        return
    
    # Check to see if the job has a publish state set. If no state is set, we move the job to uploaded. 
    # A new state needs to be created to handle jobs that are completed, but no meant to be uploaded.
    if job.publish_public != '':
         log.info(f'Processing job: {job.job_id}')
         upload_video(job=job)
    else:
        log.info("No publish state provided, marking job completed")

    if job.job_status == "error":
        log.error(f'Job {job.job_id} failed')
        jobHandler.error_job(job, "Job failed")
        return

    job.job_status = "uploaded"
    job.job_end_time = int(time())
    jobHandler.update_job(job)


def launch():

    LogHandler(service_name="videoUploader").bootstrap()
    log.info('[Boot] Starting video uploader')

    while True:
        sleep(5)
        run()


if __name__ == "__main__":

    launch()



import logging as log
from logging.handlers import TimedRotatingFileHandler
from dataclasses import dataclass
import os


@dataclass
class LogHandler:

    service_name: str = 'running'

    def __post_init__(self):
        self.dir: str = os.getenv('RWP_LOG_DIR', './')
        self.level: str = os.getenv('RWP_LOG_LEVEL', 'INFO')

    def bootstrap(self):

        logger = log.getLogger()
        logger.setLevel(getattr(log, self.level))
        formatter = log.Formatter(
            '%(asctime)s [%(levelname)s] [%(module)s] %(message)s',
            '%b %d %H:%M:%S')
        handler = TimedRotatingFileHandler(f'{self.dir}/{self.service_name}-service.log', when='midnight', interval=1, backupCount=5)

        handler.setFormatter(formatter)
        logger.addHandler(handler)

        log.info(f'[Log Bootstrap] Logging DIR set to {os.getenv("RWP_LOG_DIR", "not set")}')
        log.info(f'[Log Bootstrap] Logging to {self.dir}/{self.service_name}-service.log')



from dataclasses import dataclass, asdict

'''
A job can only be in a one discrete state at once. That state will correspond to a list queue
in Redis.

States: NEW -> IMGRESOLVED -> GRABBED -> PACKED -> UPLOADED || ERROR
'''


@dataclass
class StateManager:

    new: str = "jobs:NEW"
    imgresolved: str = "jobs:IMGRESOLVED"
    grabbed: str = "jobs:GRABBED"
    packed: str = "jobs:PACKED"
    uploaded: str = "jobs:UPLOADED"
    error: str = "jobs:ERROR"

    def __getitem__(self, item):
        return getattr(self, item)

    def valid_state(self, state: str):
        return state in asdict(self)

    def states(self):
        return asdict(self)
